//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Code to perform alpha blending and transparent BLTs for GeodeLX.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_rtns.h"
#include "cim_defs.h"
#include "cim_regs.h"
#include "blend.h"

unsigned long g_alpha;

/////////////////////////////////////////////////////////////////////////
// LXAlphaBlend
//
// Helper routine for DrvAlphaBlend.
//
void LXAlphaBlend (
    PDEV            *ppdev,
    RECTL           *prclDst,
    SURFOBJ         *psoSrc,
    ULONG            srcX,
    ULONG            srcY,
    LX_BLENDPARAMS  *blendparams)
{
    ULONG flags = 0;

    // HANDLE MULTIPLE PASS CASE
    // When blending 16BPP with 32BPP data, we cannot simply blend in the output
    // format (16BPP), as this causes mismatches in compatibility tests.  We will
    // thus use a three pass approach through the GP, hoping that it will still be
    // faster than software.
    //
    if (blendparams->threePass)
    {
        ULONG width     = prclDst->right  - prclDst->left;
        ULONG height    = prclDst->bottom - prclDst->top;
        ULONG dstOffset = (prclDst->top * blendparams->dstPitch) + (prclDst->left << ppdev->cPelSize) + blendparams->dstOffset;

        // FIRST PASS
        // Expand the destination data into 8:8:8:8 in our scratch area.
        //
        gp_declare_blt (0);
        gp_set_bpp (32);
        gp_set_source_format (CIMGP_SOURCE_FMT_0_5_6_5);
        gp_set_raster_operation (0xCC);
        gp_set_strides (width << 2, blendparams->dstPitch);
        gp_screen_to_screen_convert (
            g_fbSize, dstOffset,
            width, height, 0);

        // SECOND PASS - BLEND
        //
        gp_declare_blt (0);
        gp_set_alpha_operation (CIMGP_A_PLUS_BETA_B, CIMGP_CHANNEL_A_ALPHA,
            CIMGP_CHANNEL_A_SOURCE, CIMGP_APPLY_BLEND_TO_RGB, 0);
        gp_color_bitmap_to_screen_blt (
            g_fbSize, srcX, width, height,
            (UCHAR *)blendparams->srcOffset + (srcY * blendparams->srcPitch),
            blendparams->srcPitch);

        // THIRD PASS - CONVERT BACK
        //
        gp_declare_blt (0);
        gp_set_bpp (16);
        gp_set_raster_operation (0xCC);
        gp_set_strides (blendparams->dstPitch, width << 2);
        gp_set_source_format (CIMGP_SOURCE_FMT_8_8_8_8);
        gp_screen_to_screen_convert (dstOffset, g_fbSize, width, height, 0);

        return;
    }

    if (psoSrc->iBitmapFormat == BMF_4BPP || psoSrc->iBitmapFormat == BMF_8BPP)
        flags = CIMGP_BLTFLAGS_PRES_LUT;

    gp_declare_blt (flags);

    if (psoSrc->iBitmapFormat == BMF_1BPP)
    {
        gp_mono_bitmap_to_screen_blt (
            (prclDst->top * blendparams->dstPitch) + (prclDst->left << ppdev->cPelSize) + blendparams->dstOffset,
            srcX,
            prclDst->right  - prclDst->left,
            prclDst->bottom - prclDst->top,
            (UCHAR *)blendparams->srcOffset + (srcY * blendparams->srcPitch),
            blendparams->srcPitch);
    }
    else if (blendparams->fromFB)
    {
        // SCREEN COPY
        // Note that this code assumes that all offscreen bitmaps are in the same format as the
        // device surface.
        //
        gp_screen_to_screen_blt (
            (prclDst->top * blendparams->dstPitch) + (prclDst->left << ppdev->cPelSize) + blendparams->dstOffset,
             (srcY * blendparams->srcPitch) + (srcX << ppdev->cPelSize) + blendparams->srcOffset,
            prclDst->right  - prclDst->left,
            prclDst->bottom - prclDst->top,
            0);
    }
    else if (!blendparams->colorConversion)
    {
        gp_color_bitmap_to_screen_blt (
            (prclDst->top * blendparams->dstPitch) + (prclDst->left << ppdev->cPelSize) + blendparams->dstOffset,
            srcX,
            prclDst->right  - prclDst->left,
            prclDst->bottom - prclDst->top,
            (UCHAR *)blendparams->srcOffset + (srcY * blendparams->srcPitch),
            blendparams->srcPitch);
    }
    else
    {
        gp_color_convert_blt (
            (prclDst->top * blendparams->dstPitch) + (prclDst->left << ppdev->cPelSize) + blendparams->dstOffset,
            srcX,
            prclDst->right  - prclDst->left,
            prclDst->bottom - prclDst->top,
            (UCHAR *)blendparams->srcOffset + (srcY * blendparams->srcPitch),
            blendparams->srcPitch);
    }
}

/////////////////////////////////////////////////////////////////////////
// DrvAlphaBlend
//
// Blend source and destination data using a constant or per-pixel alpha values.
//
BOOL DrvAlphaBlend(
	SURFOBJ  *psoDst,
	SURFOBJ  *psoSrc,
	CLIPOBJ  *pco,
	XLATEOBJ *pxlo,
	RECTL    *prclDst,
	RECTL    *prclSrc,
	BLENDOBJ *pBlendObj)
{
    PDEV           *ppdev = NULL;
	RECTL           destRegion;
	RECTL          *regionPtr;
	CLIPENUM        EnumRects8;
	int             more;
    int             i;
    LX_BLENDPARAMS  blendparams;
    int             srcFormat;
    ULONG           bitfields[3];
    ULONG           index;
    RECTL           rotDestRegion;
    RECTL          *rotDestPtr;

    DISPDBG ((ppdev, 1, "DrvAlphaBlend Entry\n"));

    ACQUIRE_PRIMITIVE_SEMAPHORE;
    if (g_puntAllUntilModeSet)
        goto puntIt;

	// CHECK PUNT CASES
	// We punt if the destination surface is not in video memory or if the source
    // and destination rectangles do not have equal dimensions.
    //
	if (!psoDst->dhsurf ||
       ((prclDst->right  - prclDst->left) != (prclSrc->right  - prclSrc->left) ||
		(prclDst->bottom - prclDst->top)  != (prclSrc->bottom - prclSrc->top)))
	{
        DISPDBG ((ppdev, 1, "Punting alpha blend to offscreen surface or scaling\n"));
		goto puntIt;
	}

    ppdev = (PDEV *)psoDst->dhpdev;

    // PRE-ROTATE THE DATA
    // If the source or destination is the primary surface and we are
    // rotating, then we need to pre-rotate to our fake surface before
    // we act.  If the destination is the primary, we will also need to
    // rotate after.  It is possible for the source and destination
    // to both be the primary, in which case we *could* simply adjust
    // the dimensions and coordinates.  However, experience has shown
    // that that situation almost never occurs, if at all.
    //
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoSrc->dhsurf == 1)
    {
        rotDestPtr = prclSrc;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclSrc, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrueNoRotate;
        }

        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoDst->dhsurf == 1)
    {
        rotDestPtr = prclDst;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclDst, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrueNoRotate;
        }        
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    // PROGRAM THE SURFACE BPP
    // The driver must be able to support simultaneous rendering to multiple
    // surfaces.  The color depth of these surfaces is specified in the
    // PDEV structure.
    //
    gp_set_bpp (ppdev->cBitsPerPel);

    blendparams.colorConversion = 0;
    blendparams.alphaValue      = (BYTE)pBlendObj->BlendFunction.SourceConstantAlpha;
    blendparams.threePass      = 0;

    // CHECK FOR ALPHA IN DISPLAY MODE
    // If the display mode does not include an alpha channel, we will only apply the
    // alpha operation to the RGB components.
    //
    if (ppdev->cPelSize != 2) blendparams.alphaApply = CIMGP_APPLY_BLEND_TO_RGB;
    else                      blendparams.alphaApply = CIMGP_APPLY_BLEND_TO_ALL;

    // THREE ALPHA BLEND CASES
    // There are three different ways alpha blending can be accomplished in this
    // routine.  The source can be premultiplied by an alpha channel and those per-pixel
    // alpha values can be blended with the destination pixels.  The source and
    // destination can be multiplied with a constant alpha value.  Lastly, the
    // source can be pre-multiplied by a constant alpha and then combined using
    // per-pixel alpha with the destination pixels.  The last case requires two
    // passes per BLT.
    //
    gp_declare_blt (0);

    if (pBlendObj->BlendFunction.AlphaFormat != AC_SRC_ALPHA)
    {
#if DBG
        if (g_puntMask & LX_PUNT_CONSTANTALPHA)
            goto puntIt;
#endif

        gp_set_alpha_operation (CIMGP_ALPHA_A_PLUS_BETA_B, CIMGP_CONSTANT_ALPHA,
            CIMGP_CHANNEL_A_SOURCE, blendparams.alphaApply, blendparams.alphaValue);

        // PARSE SOURCE BITMAP FORMAT
        // Find the source bitmap format used when converting data.  Note that if the data
        // is monochrome or the same format as the display, this format will be irrelevant.
        // Also note that GeodeLX does not support all input color formats.
        //
        if (psoSrc->iBitmapFormat != psoDst->iBitmapFormat ||
            (pxlo && !(pxlo->flXlate & XO_TRIVIAL)))
        {
            blendparams.colorConversion = 1;

            switch (psoSrc->iBitmapFormat)
            {
                case BMF_1BPP:

                    gp_set_mono_source (pxlo->pulXlate[0], pxlo->pulXlate[1], 0);
                    srcFormat = 0;
                    break;

                case BMF_4BPP:

                    srcFormat = CIMGP_SOURCE_FMT_4BPP_INDEXED;
                    gp_write_parameters();
                    gp_program_lut (pxlo->pulXlate, 0);
                    gp_declare_blt (0);
                    break;

                case BMF_8BPP:

                    srcFormat = CIMGP_SOURCE_FMT_8BPP_INDEXED;
                    gp_write_parameters();
                    gp_program_lut (pxlo->pulXlate, 1);
                    gp_declare_blt (0);
                    break;

                case BMF_16BPP:

                    XLATEOBJ_cGetPalette (pxlo, XO_SRCBITFIELDS, 3, bitfields);
                    if      (bitfields[0] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_0_5_6_5;
                    else if (bitfields[0] == 0x7C00) srcFormat = CIMGP_SOURCE_FMT_1_5_5_5;
                    else if (bitfields[2] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_16BPP_BGR;
                    else                             srcFormat = CIMGP_SOURCE_FMT_15BPP_BGR;

                    break;

                case BMF_24BPP:

                    srcFormat = CIMGP_SOURCE_FMT_24BPP;
                    break;

                case BMF_32BPP:

                    srcFormat = CIMGP_SOURCE_FMT_8_8_8_8;
                    break;

                case BMF_4RLE:
                case BMF_8RLE:
                case BMF_JPEG:
                case BMF_PNG:

                    goto puntIt;

            }

            DISPDBG ((ppdev, 1, "Source format is %i\n", srcFormat));
            gp_set_source_format (srcFormat);
        }
    }

    else
    {
#if DBG
        if (g_puntMask & LX_PUNT_PREMULTIPLIED)
            goto puntIt;
#endif

        // CHECK TWO-PASS CASE
        // When SourceConstantAlpha is not 0xFF, we need to first premultiply the
        // source by a constant alpha and then blend with the destination.
        //
        if (pBlendObj->BlendFunction.SourceConstantAlpha != 0xFF)
            goto puntIt;

        // PRE-MULTIPLIED SOURCE
        // Source is guaranteed to be 32BPP ARGB.  We will set the alpha source to channel 3
        // alpha if we are doing color conversion.
        //
        else if (ppdev->cPelSize != 2)
        {
            if (((prclDst->right - prclDst->left) << 2) * (prclDst->bottom - prclDst->top) > LX_CIM_SCRATCH_SIZE)
                goto puntIt;

            blendparams.threePass = 1;
        }
        else
        {
            gp_set_alpha_operation (CIMGP_A_PLUS_BETA_B, CIMGP_CHANNEL_A_ALPHA,
                CIMGP_CHANNEL_A_SOURCE, blendparams.alphaApply, 0);
        }
    }

    // RETRIEVE SURFACE POINTER FOR PRIMARY SURFACE
    // All device bitmaps are opaque.  We must therefore extract the surface information
    // from our heap.
    //
    index = dhsurf_array[(USHORT)psoDst->dhsurf];
    if (index & CACHE_FLAG_SYSTEM)
        goto puntIt;

    blendparams.dstOffset = bitmap_heap[(USHORT)index].heap_offset;
    blendparams.dstPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;
	
    if (psoSrc->dhsurf)
    {
        index = dhsurf_array[(USHORT)psoSrc->dhsurf];
        if (index & CACHE_FLAG_SYSTEM)
        {
            blendparams.fromFB = 0;
            blendparams.srcOffset = system_heap[(USHORT)index].heap_offset;
            blendparams.srcPitch  = system_heap[(USHORT)index].flags_and_pitch;
        }
        else
        {
            blendparams.fromFB = 1;
            blendparams.srcOffset = bitmap_heap[(USHORT)index].heap_offset;
            blendparams.srcPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;
        }
    }
    else
    {
        blendparams.fromFB = 0;
        blendparams.srcOffset = (ULONG)psoSrc->pvScan0;
        blendparams.srcPitch  =  psoSrc->lDelta;
    }

    gp_set_strides (blendparams.dstPitch, blendparams.srcPitch);
    gp_write_parameters();

    if (!pco || pco->iDComplexity != DC_COMPLEX)
    {
        // FIND RELEVANT REGION

        if (pco && pco->iDComplexity == DC_RECT)
        {
            regionPtr = &destRegion;
            if (!FindIntersection (prclDst, &pco->rclBounds, regionPtr))
			    goto returnTrue;
        }
        else
            regionPtr = prclDst;

        LXAlphaBlend (ppdev, regionPtr, psoSrc,
            prclSrc->left + regionPtr->left - prclDst->left,
            prclSrc->top  + regionPtr->top  - prclDst->top,
            &blendparams);
    }

    else
	{
		// COMPLEX CLIPPED

		CLIPOBJ_cEnumStart(pco, FALSE, CT_RECTANGLES, CD_LEFTDOWN, 0);

		while (1)
		{			
			RECTL region;

			more = CLIPOBJ_bEnum(pco, sizeof (CLIPENUM), (PULONG) &EnumRects8);
														
			// ITERATE THROUGH ALL RECTANGLES IN THE CURRENT BATCH

			for (i = 0; i < EnumRects8.c; i++)
			{
				// INTERSECT WITH TARGET BOUNDS

				if (!FindIntersection (prclDst, &EnumRects8.arcl[i], &destRegion))
					continue;

                LXAlphaBlend (ppdev, &destRegion, psoSrc,
                    prclSrc->left + destRegion.left - prclDst->left,
                    prclSrc->top  + destRegion.top  - prclDst->top,
                    &blendparams);
			}

			if (!more)
				break;
		}
	}

returnTrue:

    // Rotate the result back.  The region was initialized in the pre-rotation phase.
    //
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoDst->dhsurf == 1)
    {
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (mode_pitch, rot_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            rotDestPtr->right - rotDestPtr->left,
            rotDestPtr->bottom - rotDestPtr->top,
            360 - ppdev->regSettings.RotationAngle);
    }

    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 1, "DrvAlphaBlend Exit\n"));
	return TRUE;

returnTrueNoRotate:

    // Same as return true, except that no post-rotation is performed.
    //
    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 1, "DrvAlphaBlend Exit\n"));
	return TRUE;

puntIt:

    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 1, "Punting DrvAlphaBlend\n"));
	return EngAlphaBlend(psoDst, psoSrc, pco, pxlo, prclDst, prclSrc, pBlendObj);
}

/////////////////////////////////////////////////////////////////////////
// LXTransparentBlt
//
// Helper routine for DrvTransparentBLT.
//
void LXTransparentBLT (
    PDEV       *ppdev,
    RECTL      *prclDst,
    ULONG       dstOffset,
    long        dstPitch,
    SURFOBJ    *psoSrc,
    ULONG       srcOffset,
    long        srcPitch,
    ULONG       srcX,
    ULONG       srcY,
    ULONG       iTransColor,
    int         srcFormat,
    int         colorConversion,
    int         fromFB)
{
    ULONG   width;
    ULONG   stride;
    ULONG   height;

    if (!colorConversion)
    {
        // NO COLOR CONVERSION - SIMPLE COPY
        //
        gp_declare_blt (0);

        if (fromFB)
        {
            gp_screen_to_screen_blt (
                (prclDst->top * dstPitch) + (prclDst->left << ppdev->cPelSize) + dstOffset,
                (srcY * srcPitch) + (srcX << ppdev->cPelSize) + srcOffset,
                prclDst->right  - prclDst->left,
                prclDst->bottom - prclDst->top,
                0);
        }
        else
        {
            gp_color_bitmap_to_screen_blt (
                (prclDst->top * dstPitch) + (prclDst->left << ppdev->cPelSize) + dstOffset,
                srcX,
                prclDst->right  - prclDst->left,
                prclDst->bottom - prclDst->top,
                (UCHAR *)srcOffset + (srcY * srcPitch),
                srcPitch);
        }
    }
    else
    {
        // COLOR CONVERSION - TWO OR THREE PASSES
        //
        if (srcFormat == CIMGP_SOURCE_FMT_8_8_8_8)
        {
            width  = prclDst->right  - prclDst->left;
            height = prclDst->bottom - prclDst->top;
            stride = width << 2;

            // 8:8:8:8 TO 5:6:5 - THREE PASSES
            // On the first pass, convert the 5:6:5 destination data into 8:8:8:8 inside
            // the temporary buffer.
            //
            gp_declare_blt (0);
            gp_set_bpp (32);
            gp_set_raster_operation (0xCC);
            gp_set_source_format (CIMGP_SOURCE_FMT_0_5_6_5);
            gp_set_strides (stride, dstPitch);
            gp_screen_to_screen_convert (
                g_fbSize,
                (prclDst->top * dstPitch) + (prclDst->left << ppdev->cPelSize) + dstOffset,
                width, height, 0);

            // SECOND PASS - COPY THE DATA WITH TRANSPARENCY
            // Note that the ROP and BPP remain from the previous BLT.
            //
            gp_declare_blt (0);
            gp_set_source_transparency (iTransColor, 0xFFFFFFFF);
            gp_color_bitmap_to_screen_blt (
                g_fbSize, srcX, width, height,
                (UCHAR *)srcOffset + (srcY * srcPitch),
                srcPitch);

            // FINAL PASS - BLT THE DATA BACK
            //
            gp_declare_blt (0);
            gp_set_bpp (16);
            gp_set_strides (dstPitch, stride);
            gp_set_raster_operation (0xCC);
            gp_set_source_format (CIMGP_SOURCE_FMT_8_8_8_8);
            gp_screen_to_screen_convert (
                (prclDst->top * dstPitch) + (prclDst->left << ppdev->cPelSize) + dstOffset,
                g_fbSize, width, height, 0);
        }
        else
        {
            width  = prclDst->right  - prclDst->left;
            height = prclDst->bottom - prclDst->top;

            if (ppdev->cPelSize == 2)
            {
                stride = width << 2;

                // 1:5:5:5 OR 5:6:5 TO 8:8:8:8 - TWO PASSES
                // On the first pass, expand the data into the temporary buffer.
                //
                gp_declare_blt(0);
                gp_set_raster_operation (0xCC);
                gp_set_source_format (srcFormat);
                gp_set_strides (stride, 0);
                gp_color_convert_blt (
                    g_fbSize,
                    srcX,
                    width, height,
                    (UCHAR *)srcOffset + (srcY * srcPitch),
                    srcPitch);

                // SECOND PASS
                // Copy the data to the frame buffer using the expanded color key
                //
                gp_declare_blt (0);
                gp_set_strides (dstPitch, stride);
                if (srcFormat == CIMGP_SOURCE_FMT_1_5_5_5)
                {
                    gp_set_source_transparency (
                        (((iTransColor & 0x7C00) >> 7) | ((iTransColor & 0x7000) >> 12) << 16) |
                        (((iTransColor & 0x03E0) >> 2) | ((iTransColor & 0x0380) >>  7) << 8)  |
                        (((iTransColor & 0x001F) << 3) | ((iTransColor & 0x001C) >>  2)),
                        0xFFFFFF);
                }
                else
                {
                    gp_set_source_transparency (
                        (((iTransColor & 0xF800) >> 8) | ((iTransColor & 0xE000) >> 13) << 16) |
                        (((iTransColor & 0x07E0) >> 3) | ((iTransColor & 0x0600) >>  9) << 8)  |
                        (((iTransColor & 0x001F) << 3) | ((iTransColor & 0x001C) >>  2)),
                        0xFFFFFF);
                }
                gp_screen_to_screen_blt (
                    (prclDst->top * dstPitch) + (prclDst->left << ppdev->cPelSize) + dstOffset,
                    g_fbSize,
                    width, height, 0);
            }

            else
            {
                // 1:5:5:5 TO 5:6:5
                // Similar to above except that the temp buffer is 16BPP
                //
                stride = ((width << 1) + 2) & 0xFFFC;

                gp_declare_blt(0);
                gp_set_raster_operation (0xCC);
                gp_set_source_format (srcFormat);
                gp_set_strides (stride, 0);
                gp_color_convert_blt (
                    g_fbSize,
                    srcX,
                    width, height,
                    (UCHAR *)srcOffset + (srcY * srcPitch),
                    srcPitch);

                // SECOND PASS
                // Copy the data to the frame buffer using the expanded color key
                //
                gp_declare_blt (0);
                gp_set_strides (dstPitch, stride);
                gp_set_source_transparency (
                    ((iTransColor << 1) & 0xFFC0) |
                    ((iTransColor >> 5) & 0x0020) |
                     (iTransColor & 0x1F),
                     0xFFFF);
                gp_screen_to_screen_blt (
                    (prclDst->top * dstPitch) + (prclDst->left << ppdev->cPelSize) + dstOffset,
                    g_fbSize,
                    width, height, 0);
            }
        }
    }
}

/////////////////////////////////////////////////////////////////////////
// DrvTransparentBlt
//
// Source copy using transparency
//
BOOL DrvTransparentBlt(
	SURFOBJ  *psoDst,
	SURFOBJ  *psoSrc,
	CLIPOBJ  *pco,
	XLATEOBJ *pxlo,
	RECTL    *prclDst,
	RECTL    *prclSrc,
	ULONG     iTransColor,
	ULONG     ulReserved)
{
	PDEV       *ppdev = NULL;
	RECTL       destRegion;
	RECTL      *regionPtr;
	CLIPENUM    EnumRects8;
	int         more;
    int         i;
    int         srcFormat;
    int         colorConversion = 0;
    ULONG       dstOffset, srcOffset;
    long        dstPitch,  srcPitch;
    ULONG       bitfields[3];
    ULONG       passes;
    ULONG       index, fromFB;
    RECTL       rotDestRegion;
    RECTL      *rotDestPtr;

    DISPDBG ((ppdev, 1, "DrvTransparentBlt Entry\n"));

    ACQUIRE_PRIMITIVE_SEMAPHORE;
    if (g_puntAllUntilModeSet)
        goto puntIt;

	if (!psoDst->dhsurf ||
       ((prclDst->right  - prclDst->left) != (prclSrc->right  - prclSrc->left) ||
		(prclDst->bottom - prclDst->top)  != (prclSrc->bottom - prclSrc->top)))
	{
        DISPDBG ((ppdev, 1, "Punting transparent BLT to offscreen surface or scaling\n"));
		goto puntIt;
	}

	ppdev = (PDEV *)psoDst->dhpdev;

    // PRE-ROTATE THE DATA
    // If the source or destination is the primary surface and we are
    // rotating, then we need to pre-rotate to our fake surface before
    // we act.  If the destination is the primary, we will also need to
    // rotate after.  It is possible for the source and destination
    // to both be the primary, in which case we *could* simply adjust
    // the dimensions and coordinates.  However, experience has shown
    // that that situation almost never occurs, if at all.
    //
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoSrc->dhsurf == 1)
    {
        rotDestPtr = prclSrc;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclSrc, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrueNoRotate;
        }

        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoDst->dhsurf == 1)
    {
        rotDestPtr = prclDst;
        if (pco && pco->iDComplexity != DC_TRIVIAL)
        {
            rotDestPtr = &rotDestRegion;
            if (!FindIntersection (prclDst, &ppdev->scrnBounds, rotDestPtr))
                goto returnTrueNoRotate;
        }        
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (rot_pitch, mode_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_WIDTH (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ROT_HEIGHT (rotDestPtr->, ppdev->regSettings.RotationAngle),
            ppdev->regSettings.RotationAngle);
    }

    // PROGRAM THE SURFACE BPP
    // The driver must be able to support simultaneous rendering to multiple
    // surfaces.  The color depth of these surfaces is specified in the
    // PDEV structure.
    //
    gp_set_bpp (ppdev->cBitsPerPel);

	if (psoSrc->iBitmapFormat != ppdev->iBitmapFormat ||
		(pxlo && !(pxlo->flXlate & XO_TRIVIAL)))
	{
#if DBG
        if (g_puntMask & LX_PUNT_TRANSPARENTCONVERT)
            goto puntIt;
#endif

        colorConversion = 1;

        // PARSE SOURCE BITMAP FORMAT
        // Due to limitations in the GeodeLX transparency hardware, we only support the
        // following conversions: 5:6:5 -> 8:8:8:8, 1:5:5:5 -> 8:8:8:8,
        // 1:5:5:5 -> 5:6:5, and 8:8:8:8 -> 5:6:5.  The first three can be accomplished
        // using a two pass approach.  The first pass converts the data and copies it to a
        // temporary buffer.  The second pass copies the data using the expanded color
        // key.  As our destination depth has more bits than the source, we have no
        // aliasing on the color key.  The last conversion case must be accomplished
        // using three passes.  The first pass copies the destination data into a temporary
        // buffer.  The second copies the data using transparency on top of the temp buffer.
        // The last pass converts the data back into the dest format.  No data is lost
        // here because the conversion from 5:6:5 to 8:8:8:8 and back again does not change
        // any bits.  With two to three passes per, why are we simply not punting?  To avoid
        // the pipeline stall that punts inevitably require.
        //
        if (psoSrc->iBitmapFormat == BMF_32BPP)
        {
            // CHECK FOR 8:8:8:8 TO 5:6:5
            //
            if (psoDst->iBitmapFormat != BMF_16BPP)
            {
                DISPDBG ((ppdev, 1000, "Punting 32BPP transparent BLT\n"));
                goto puntIt;
            }

            // VERIFY THE DATA WILL FIT INSIDE THE TEMP BUFFER
            //
            if ((prclDst->bottom - prclDst->top) * ((prclDst->right - prclDst->left) << 2) > LX_CIM_SCRATCH_SIZE)
            {
                DISPDBG ((ppdev, 1000, "32BPP transparent BLT too large\n"));
                goto puntIt;
            }

            srcFormat = CIMGP_SOURCE_FMT_8_8_8_8;
        }
        else if (psoSrc->iBitmapFormat == BMF_16BPP)
        {
            XLATEOBJ_cGetPalette (pxlo, XO_SRCBITFIELDS, 3, bitfields);
            if      (bitfields[0] == 0xF800) srcFormat = CIMGP_SOURCE_FMT_0_5_6_5;
            else if (bitfields[0] == 0x7C00) srcFormat = CIMGP_SOURCE_FMT_1_5_5_5;
            else
            {
                DISPDBG ((ppdev, 1000, "Punting 16BPP transparent BLT\n"));
                goto puntIt;
            }

            // CHECK FOR AVAILABLE SPACE
            //
            if ((prclDst->bottom - prclDst->top) * (((prclDst->right - prclDst->left) << ppdev->cPelSize) + 4) > LX_CIM_SCRATCH_SIZE)
            {
                DISPDBG ((ppdev, 1000, "16BPP Transparent BLT too large\n"));
                goto puntIt;
            }
        }
        else
        {
            DISPDBG ((ppdev, 1000, "Punting color converted transparent BLT - %i\n", psoSrc->iBitmapFormat));
            goto puntIt;
        }
    }
    else
    {
#if DBG
        if (g_puntMask & LX_PUNT_TRANSPARENTCOPY)
            goto puntIt;
#endif
    }

    // RETRIEVE SURFACE POINTER FOR PRIMARY SURFACE
    // All device bitmaps are opaque.  We must therefore extract the surface information
    // from our heap.
    //
    index = dhsurf_array[(USHORT)psoDst->dhsurf];
    if (index & CACHE_FLAG_SYSTEM)
        goto puntIt;

    dstOffset = bitmap_heap[(USHORT)index].heap_offset;
    dstPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;
	
    if (psoSrc->dhsurf)
    {
        index = dhsurf_array[(USHORT)psoSrc->dhsurf];
        if (index & CACHE_FLAG_SYSTEM)
        {
            fromFB = 0;
            srcOffset = system_heap[(USHORT)index].heap_offset;
            srcPitch  = system_heap[(USHORT)index].flags_and_pitch;
        }
        else
        {
            fromFB = 1;
            srcOffset = bitmap_heap[(USHORT)index].heap_offset;
            srcPitch  = bitmap_heap[(USHORT)index].flags_and_pitch;
        }
    }
    else
    {
        fromFB = 0;
        srcOffset = (ULONG)psoSrc->pvScan0;
        srcPitch  = psoSrc->lDelta;
    }

    if (!colorConversion)
    {
        gp_declare_blt (0);
        gp_set_raster_operation (0xCC);
        gp_set_source_transparency (iTransColor, 0xFFFFFFFF);
        gp_set_strides (dstPitch, srcPitch);
        gp_write_parameters();
    }

    if (!pco || pco->iDComplexity != DC_COMPLEX)
    {
        // FIND RELEVANT REGION

        if (pco && pco->iDComplexity == DC_RECT)
        {
            regionPtr = &destRegion;
            if (!FindIntersection (prclDst, &pco->rclBounds, regionPtr))
			    goto returnTrue;
        }
        else
            regionPtr = prclDst;

        LXTransparentBLT (ppdev, regionPtr, dstOffset, dstPitch, psoSrc, srcOffset, srcPitch,
            prclSrc->left + regionPtr->left - prclDst->left,
            prclSrc->top  + regionPtr->top  - prclDst->top,
            iTransColor, srcFormat, colorConversion, fromFB);
    }

    else
	{
		// COMPLEX CLIPPED

		CLIPOBJ_cEnumStart(pco, FALSE, CT_RECTANGLES, CD_LEFTDOWN, 0);

		while (1)
		{			
			RECTL region;

			more = CLIPOBJ_bEnum(pco, sizeof (CLIPENUM), (PULONG) &EnumRects8);
														
			// ITERATE THROUGH ALL RECTANGLES IN THE CURRENT BATCH

			for (i = 0; i < EnumRects8.c; i++)
			{
				// INTERSECT WITH TARGET BOUNDS

				if (!FindIntersection (prclDst, &EnumRects8.arcl[i], &destRegion))
					continue;

                LXTransparentBLT (ppdev, &destRegion, dstOffset, dstPitch, psoSrc, srcOffset, srcPitch,
                    prclSrc->left + destRegion.left - prclDst->left,
                    prclSrc->top  + destRegion.top  - prclDst->top,
                    iTransColor, srcFormat, colorConversion, fromFB);		
			}

			if (!more)
				break;
		}
	}

returnTrue:

    // Rotate the result back.  The region was initialized in the pre-rotation phase.
    //
    if (ppdev->regSettings.RotationAngle != 0 && (USHORT)psoDst->dhsurf == 1)
    {
        gp_set_bpp (ppdev->cBitsPerPel);
        gp_declare_blt (0);
        gp_set_strides (mode_pitch, rot_pitch);
        gp_set_source_format (ppdev->cimFormat);
        gp_rotate_blt (
            ROT_OFFSET (ppdev, rotDestPtr->, ppdev->regSettings.RotationAngle),
            (rot_pitch * rotDestPtr->top) + (rotDestPtr->left << ppdev->cPelSize) + rot_offset,
            rotDestPtr->right - rotDestPtr->left,
            rotDestPtr->bottom - rotDestPtr->top,
            360 - ppdev->regSettings.RotationAngle);
    }

    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 0, "DrvTransparentBlt Exit\n"));
	return TRUE;

returnTrueNoRotate:

    // Same as return true, except that no post-rotation is performed.
    //
    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 1, "DrvTransparentBlt Exit\n"));
	return TRUE;

puntIt:

    RELEASE_PRIMITIVE_SEMAPHORE;

    DISPDBG ((ppdev, 0, "Punting DrvTransparentBlt\n"));
	return EngTransparentBlt (psoDst, psoSrc, pco, pxlo, prclDst,
		prclSrc, iTransColor, ulReserved);
}
