//
// <LIC_AMD_STD>
// Copyright (C) <years> Advanced Micro Devices, Inc.  All Rights Reserved.
// </LIC_AMD_STD>
//
// <CTL_AMD_STD>
// </CTL_AMD_STD>
//
// <DOC_AMD_STD>
// Brush Realization and creation via DrvRealizeBrush.
// </DOC_AMD_STD>
//

#include "precomp.h"
#include "cim_parm.h"

/////////////////////////////////////////////////////////////////////////
// DrvRealizeBrush
//
// Translate brushes into a form that the driver can use.
//
BOOL DrvRealizeBrush(
	BRUSHOBJ * pbo,
	SURFOBJ * psoTarget,
	SURFOBJ * psoPattern,
	SURFOBJ * psoMask,
	XLATEOBJ * pxlo,
	ULONG iHatch)
{
	PDEV       *ppdev = NULL;
    PPBRUSH     pBrush;
    ULONG       bitfields[3];
    ULONG       diff;
    ULONG       i;
    ULONG      *pData;
    ULONG      *pSrc;

	DISPDBG ((ppdev, 0, "DrvRealizeBrush Entry\n"));

	// WE SUPPORT ONLY 8X8 BRUSHES
    // In truth we could support 2x2 and 4x4 brushes, but 8x8 seems to be
    // the standard brush size.  Note also that previous incarnations of this
    // driver assumed the type of the SURFOBJ was STYPE_BITMAP.  However,
    // as our device bitmaps are non-opaque, we can assume that a pitch and
    // delta exist for incoming surfaces, regardless of the surface type.
    //
	if((psoPattern->sizlBitmap.cx != 8) || (psoPattern->sizlBitmap.cy != 8))
    {
		DISPDBG((ppdev, 1, "Can only handle 8x8 bitmaps, not %ix%i\n", psoPattern->sizlBitmap.cx,
            psoPattern->sizlBitmap.cy));
		return FALSE;
	}

    if (psoMask)
    {
        DISPDBG ((ppdev, 1, "Punting masked pattern\n"));
        return FALSE;
    }

    // DECODE SURFACE FORMAT
    // We are currently only supporting monochrome, 15BPP, 16BPP and 32BPP.
    //
    if (psoPattern->iBitmapFormat == BMF_1BPP)
    {
        if ((pBrush = (PPBRUSH)BRUSHOBJ_pvAllocRbrush (pbo, sizeof(PBRUSH) + 8)) == NULL)
        {
            DISPDBG ((ppdev, 1, "Failed mono brush allocation\n"));
            return FALSE;
        }
        pBrush->patFormat  = 0xFF;
        pBrush->monoColor0 = pxlo->pulXlate[0];
        pBrush->monoColor1 = pxlo->pulXlate[1];

        // CHECK FOR CONTIGUOUS DATA
        //
        if (psoPattern->lDelta == 1)
        {
            pBrush->Pattern[0] = *(ULONG *)(psoPattern->pvScan0);
            pBrush->Pattern[1] = *(ULONG *)((ULONG)psoPattern->pvScan0 + 4);
        }
        else
        {
            for (i = 0; i < 8; i++)
            {
                ((BYTE *)(&pBrush->Pattern[0]))[i] =
                    *((BYTE *)((ULONG)psoPattern->pvScan0 + (i * psoPattern->lDelta)));
            }
        }
    }
    else if (psoPattern->iBitmapFormat == BMF_16BPP)
    {
        if ((pBrush = (PPBRUSH)BRUSHOBJ_pvAllocRbrush (pbo, sizeof(PBRUSH) + 128)) == NULL)
        {
            DISPDBG ((ppdev, 1, "Failed 16BPP brush allocation\n"));
            return FALSE;
        }

        // DECODE THE SOURCE FORMAT
        // Brushes may be 15BPP or 16BPP
        //
        if (pxlo && pxlo->flXlate != XO_TRIVIAL)
        {
            XLATEOBJ_cGetPalette (pxlo, XO_SRCBITFIELDS, 3, bitfields);
            if      (bitfields[0] == 0xF800) pBrush->patFormat = CIMGP_SOURCE_FMT_0_5_6_5;
            else if (bitfields[0] == 0x7C00) pBrush->patFormat = CIMGP_SOURCE_FMT_1_5_5_5;
            else
            {
                DISPDBG ((ppdev, 1, "Unsupported brush type\n"));
                return FALSE;
            }
        }
        else
            pBrush->patFormat = CIMGP_SOURCE_FMT_0_5_6_5;

        pSrc  = (ULONG *)psoPattern->pvScan0;
        pData = &(pBrush->Pattern[0]);

        // CHECK FOR CONTIGUOUS DATA

        if (psoPattern->lDelta == 16)
        {
            _asm {
                cld
                mov     edi, pData
                mov     esi, pSrc
                mov     ecx, 32
                rep     movsd
            }
        }
        else
        {
            diff = (ULONG)psoPattern->lDelta - 16;
            _asm {
                cld
                mov     edi, pData
                mov     esi, pSrc
                mov     edx, 8
loop16:
                mov     ecx, 4
                rep     movsd
                add     esi, diff
                dec     edx
                jnz     loop16
            }
        }
    }
    else if (psoPattern->iBitmapFormat == BMF_32BPP)
    {
        if ((pBrush = (PPBRUSH)BRUSHOBJ_pvAllocRbrush (pbo, sizeof(PBRUSH) + 256)) == NULL)
        {
            DISPDBG ((ppdev, 1, "Failed 32BPP brush allocation\n"));
            return FALSE;
        }

        pBrush->patFormat = CIMGP_SOURCE_FMT_8_8_8_8;

        pSrc  = (ULONG *)psoPattern->pvScan0;
        pData = &(pBrush->Pattern[0]);

        // CHECK FOR CONTIGUOUS DATA

        if (psoPattern->lDelta == 32)
        {
            _asm {
                cld
                mov     edi, pData
                mov     esi, pSrc
                mov     ecx, 64
                rep     movsd
            }
        }
        else
        {
            diff = (ULONG)psoPattern->lDelta - 32;
            _asm {
                cld
                mov     edi, pData
                mov     esi, pSrc
                mov     edx, 8
loop32:
                mov     ecx, 8
                rep     movsd
                add     esi, diff
                dec     edx
                jnz     loop32
            }
        }
    }
    else
    {
        DISPDBG ((ppdev, 1, "Unsupported brush type.\n"));
        return FALSE;
    }

    DISPDBG ((ppdev, 0, "DrvRealizeBrush Exit\n"));
    return TRUE;
}
