// Copyright (c) 2006 Advanced Micro Devices ("AMD").
//
// All Rights Reserved.  Unpublished rights reserved under the 
// copyright laws of the United States of America, other countries 
// and international treaties.  The software is provided without 
// fee.  Permission to use, copy, store, modify, disclose, transmit 
// or distribute the software is granted, provided that this 
// copyright notice must appear in any copy, modification, 
// disclosure, transmission or distribution of the software.
//
// AMD retains all ownership, copyright, trade secret and 
// proprietary rights in the software. THIS SOFTWARE HAS BEEN 
// PROVIDED "AS IS," WITHOUT EXPRESS OR IMPLIED WARRANTY INCLUDING, 
// WITHOUT LIMITATION, IMPLIED WARRANTIES OF MERCHANTABILITY, 
// FITNESS FOR A PARTICULAR USE AND NON-INFRINGEMENT.
//
// AMD does not assume or authorize any other person to assume for 
// it any liability in connection with the Software. AMD SHALL NOT 
// BE LIABLE TO COMPANY, OR ANY THIRD PARTY, IN CONTRACT, TORT, 
// WARRANTY, STRICT LIABILITY, OR OTHERWISE FOR ANY DIRECT DAMAGES, 
// OR FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR CONSEQUENTIAL 
// DAMAGES, INCLUDING BUT NOT LIMITED TO, BUSINESS INTERRUPTION, 
// LOST PROFITS OR GOODWILL, OR LOSS OF INFORMATION EVEN IF AMD IS 
// ADVISED OF THE POSSIBILITY OF SUCH DAMAGES.
//
#include <stdlib.h>

// MINIMA AND MAXIMA, IN MHZ

#define MIN_VCO_FREQ      200.0
#define MAX_VCO_FREQ      450.0
#define DEFAULT_FREQ_REF  48.0
#define MAX_N             255
#define MAX_M             7
#define MAX_P             15
#define MAX_PLL_COUNT     100

#define PLL_LIST_TYPE_NORMAL      0
#define PLL_LIST_TYPE_INVALID     1

#define SORT_BY_TOLERANCE    0
#define SORT_BY_LOWM         1

#define MAG(a)   (((a) < 0) ? -(a) : (a))
#define MIN(a,b) (((a) < (b)) ? (a) : (b))

typedef struct tagpllfreq
{
	double frequency;
	double tolerance;
	int m;
	int n;
	int p;
    int list_type;

} PLL_FREQ;

typedef struct tagtestparams
{
	double frequency;
	double tolerance;
	double freq_ref;
	int freq_count;
	int sort_mode;
	int display_list;
    int include_invalid;

} TEST_PARAMS;

// PLL FREQUENCY ACCUMULATOR

PLL_FREQ frequencies[MAX_PLL_COUNT];

// ARRAY SORT OPTIONS

int sort_type = SORT_BY_TOLERANCE;
int num_valid = 0;
int CalculatePllValues (TEST_PARAMS *test_params);
int ComparePllFrequencies (const void *elem1, const void*elem2);

unsigned long CalculatePLL (double frequency)
{
	TEST_PARAMS test_params;

	// INITIALIZE TEST DEFAULTS

	test_params.tolerance       = 0;
	test_params.freq_ref        = DEFAULT_FREQ_REF;
	test_params.display_list    = 0;
	test_params.sort_mode       = SORT_BY_TOLERANCE;
	test_params.freq_count      = MAX_PLL_COUNT;
    test_params.include_invalid = 0;
    test_params.frequency       = frequency;
	
	CalculatePllValues (&test_params);
    qsort ((void *)frequencies, MAX_PLL_COUNT, sizeof (PLL_FREQ), ComparePllFrequencies);

	return frequencies[0].p | (frequencies[0].n << 4) | (frequencies[0].m << 12);
}

//---------------------------------
// PLL Calculation Routines
//---------------------------------

//******************************************************************
// FindWorstToleranceIndex
//
// This function returns the array index containing the highest
// tolerance.
//------------------------------------------------------------------

int FindWorstToleranceIndex (PLL_FREQ *pll_array, int max_count)
{
	int i, index = 0;
	double max_tolerance = 0;

	for (i = 0; i < max_count; i++)
	{
		if (pll_array[i].tolerance > max_tolerance)
		{
			index = i;
			max_tolerance = pll_array[i].tolerance;
		}
	}

	return index;
}

//******************************************************************
// AddPllEntry
//   
// This function adds an entry to the PLL array in place of the
//entry with the highest tolerance.
//------------------------------------------------------------------

double AddPllEntry (PLL_FREQ *pll_array, int max_count, double frequency, 
					 double tolerance, int m, int n, int p, int invalid)
{
	int index;
	
	// ADD ENTRY IN PLACE OF WORST TOLERANCE

	index = FindWorstToleranceIndex (pll_array, max_count);

	pll_array[index].tolerance = tolerance;
	pll_array[index].frequency = frequency;
	pll_array[index].m = m;
	pll_array[index].n = n;
	pll_array[index].p = p;
    pll_array[index].list_type = invalid ? PLL_LIST_TYPE_INVALID : PLL_LIST_TYPE_NORMAL;
	
	// FIND NEW WORST TOLERANCE

	index = FindWorstToleranceIndex (pll_array, max_count);

	return pll_array[index].tolerance;
}

//*******************************************************************
// CalculatePllValues
//
// This function creates an array of the 50 closest PLL frequency
// matches.  Less than 50 of these matches may meet the tolerance
// requirement.
//------------------------------------------------------------------

int CalculatePllValues (TEST_PARAMS *test_params)
{
	int m, n, p, i;
    int invalid;
	double frequency, tolerance, pll_count, freq_ref;
	double worst_tolerance, current_tolerance;
	double current_frequency;
    double vco;
	
	// READ PLL PARAMETERS

	tolerance = test_params->tolerance;
	frequency = test_params->frequency;
	freq_ref  = test_params->freq_ref;
	pll_count = test_params->freq_count;
	
	// INITIALIZE ALL TOLERANCES TO MAXIMUM

	for (i = 0; i < MAX_PLL_COUNT; i++)
    {
		frequencies[i].tolerance = 10000.0;
        frequencies[i].list_type = PLL_LIST_TYPE_NORMAL;
    }

	worst_tolerance = 10000.0;

	if (frequency > MAX_VCO_FREQ && !test_params->include_invalid)
	{
		return 0;
	}

	// CREATE ARRAY OF LOWEST TOLERANCES

    for (m = 0; m < MAX_M; m++)
    {
        for (n = 0; n < MAX_N; n++)
        {
            vco = (freq_ref * (double)(n + 1)) / (double)(m + 1); 

            if (vco > MAX_VCO_FREQ || vco < MIN_VCO_FREQ)
            {
                invalid = 1;
                if (!test_params->include_invalid)
                    continue;
            }
            else
                invalid = 0;

            for (p = 0; p < MAX_P; p++)
            {
                current_frequency  = freq_ref * (double)(n + 1);
                current_frequency /= (double)((m + 1) * (p + 1));
                current_tolerance = (MAG (current_frequency - frequency)) / frequency;

                if (current_tolerance < worst_tolerance)
			    {
				    // ADD ENTRY TO ARRAY

				    worst_tolerance = AddPllEntry (frequencies, MAX_PLL_COUNT, current_frequency,
					    current_tolerance, m, n, p, invalid);
    				
				    if (current_tolerance <= tolerance)
					    num_valid++;
			    }
            }
        }
    }

	return 0;
}

//*******************************************************************
// ComparePllFrequencies
//
// Function used in calls to qsort.  This function sorts elements
// of the frequencies array in one of several ways.
//------------------------------------------------------------------

int ComparePllFrequencies (const void *elem1, const void*elem2)
{
	PLL_FREQ *freq1 = (PLL_FREQ *)elem1;
	PLL_FREQ *freq2 = (PLL_FREQ *)elem2;
		
	// SET SORTING ORDER
	// When sorting by Tolerance: elements are sorted by tolerance and then low m
	// When sorting by Lowest M : elements are sorted by lowest M, then tolerance

    if (sort_type == SORT_BY_TOLERANCE)
    {
        if (freq1->tolerance != freq2->tolerance)
            return ((freq1->tolerance < freq2->tolerance) ? -1 : 1);
        if (freq1->m != freq2->m)
			return ((freq1->m < freq2->m) ? -1 : 1);
    }
    else
    {
        if (freq1->m != freq2->m)
			return ((freq1->m < freq2->m) ? -1 : 1);
        if (freq1->tolerance != freq2->tolerance)
            return ((freq1->tolerance < freq2->tolerance) ? -1 : 1);
    }

    return 0;
}

